unit uPanel;
{$INCLUDE 'NamedFoldersLite.inc'}

interface

uses
  Windows, SysUtils, Classes,
  {$IFDEF FAR3} Plugin3, {$ELSE} PluginW, {$ENDIF} FarKeysW, FarColor,
  NativeXml,
  uLog, uFAR, uOptions, uMessages, uConsts, uXml, uFunctions,
  uConfirmDialog, uInfo;

type
  TNamedFoldersPanel = class
  public
    type
      TOverwriteState = (osDefault, osAlwaysOverwrite, osAlwaysSkip);
  protected
    type
      TPanelItemInfo = record
        FileName: string;
        PanelFileName: string;
        Description: string;
        IsDirectory: boolean;
        //Flags: integer;
      end;
      TPanelItemArray = array of TPanelItemInfo;
  private
    fPanelItems: packed array of TPluginPanelItem;
    fPanelItemsHelp: TPanelItemArray;
    fPanelItemsValid: boolean;
    fDirectory: string;
    fTitle: string;
    fFileToBeSelected: string;
    function GetFarDirectory: PFarChar;
    function GetFarTitle: PFarChar;
    procedure SetDirectory(const Value: string);
    procedure SaveLastDirectory;
    procedure OpenLastDirectory;
    function GetHandle: THandle;
  protected
    procedure ValidatePanelItems;
    procedure ReloadPanelContents;
    function DecodeCanOverwriteResult(Res: TConfirmOverwriteDialog.TConfirmOverwrite; var OverwriteState: TOverwriteState): boolean;
    function AskCanOverwriteNamedFolder(OldInfo, NewInfo: TNamedFolderInfo; var OverwriteState: TOverwriteState): boolean;
    function AskCanOverwriteFile(const FileName: string; var OverwriteState: TOverwriteState): boolean;
  public
    class function HandleToPanel(Handle: THandle; out Panel: TNamedFoldersPanel): boolean;
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function GetPanelItems: PPluginPanelItemArray;
    function GetPanelItemsCount: integer;
    function DeleteNamedFolder(const Name: string): boolean;
    function EditNamedFolder(const Name: string): boolean;
    function NewNamedFolder: boolean;
    function ExecuteNamedFolder(InCurrentPanel: boolean; out Handled: boolean): boolean;
    function ExportNamedFolderToDirectory(const Name, TargetDirectory: string; DeleteOriginal, Silent: boolean; var OverwriteState: TOverwriteState): boolean;
    function ExportNamedFolderToFile(const Name, TargetFileName: string; DeleteOriginal, Silent: boolean; var OverwriteState: TOverwriteState): boolean;
    function ImportNamedFolderFile(const FileName, SourceDirectory: string; DeleteOriginal, Silent: boolean; var OverwriteState: TOverwriteState): boolean;
    function MakeDirectory(const Name: string): boolean;
    function ChangeDirectory(const Name: string): boolean;
    function SelectFile(const Name: string): boolean;
    property Handle: THandle read GetHandle;
    property Directory: string read fDirectory;
    property FarDirectory: PFarChar read GetFarDirectory;
    property FarTitle: PFarChar read GetFarTitle;
    property FileToBeSelected: string read fFileToBeSelected write fFileToBeSelected;
  end;

implementation

uses
  uTasks;

const
  XML_ROOTNODE = 'named_folders_lite';
  XML_ITEMNODE = 'item';

{ TNamedFoldersPanel }

function TNamedFoldersPanel.AskCanOverwriteFile(const FileName: string; var OverwriteState: TOverwriteState): boolean;
begin
  case OverwriteState of
    osDefault:
      Result := DecodeCanOverwriteResult(TConfirmOverwriteFileDialog.ShowModal(FileName), OverwriteState);
    osAlwaysOverwrite:
      Result := True;
    osAlwaysSkip:
      Result := False;
    else
      Result := False;
  end;
end;

function TNamedFoldersPanel.AskCanOverwriteNamedFolder(OldInfo, NewInfo: TNamedFolderInfo; var OverwriteState: TOverwriteState): boolean;
begin
  case OverwriteState of
    osDefault:
      Result := DecodeCanOverwriteResult(TConfirmOverwriteNamedFolderDialog.ShowModal(OldInfo, NewInfo), OverwriteState);
    osAlwaysOverwrite:
      Result := True;
    osAlwaysSkip:
      Result := False;
    else
      Result := False;
  end;
end;

function TNamedFoldersPanel.ChangeDirectory(const Name: string): boolean;
var
  FullName: string;
  Info: TNamedFolderInfo;
begin
  //Log('TNamedFoldersPanel.ChangeDirectory("%s")', [Name]);
  Result := False;
  FullName := GetAbsoluteDirectory(Self.Directory, Name);
  //Log('- absolute dir "%s"', [FullName]);
  if FullName = '' then begin
    //Log('- going to root');
    SetDirectory('/');
    Result := True;
  end
  else if TNamedFolders.ReadInfoFromOptions(FullName, Info, False) then
    try
      //Log('- found info');
      if Info.IsDirectory then begin
        //Log('going to "/%s"', [FullName]);
        SetDirectory('/' + FullName);
        Result := True;
      end;
    finally
      FreeAndNil(Info);
    end
  else begin
    //Log('- directory not found');
  end;
  //Log('TNamedFoldersPanel.ChangeDirectory("%s") -> %d', [Name, Integer(Result)]);
end;

procedure TNamedFoldersPanel.Clear;
begin
  fPanelItemsValid := False;
end;

constructor TNamedFoldersPanel.Create;
begin
  inherited Create;
  OpenLastDirectory;
end;

function TNamedFoldersPanel.DecodeCanOverwriteResult(Res: TConfirmOverwriteDialog.TConfirmOverwrite; var OverwriteState: TOverwriteState): boolean;
begin
  Result := False;
  case Res of
    coOverwrite:
      begin
        Result := True;
      end;
    coOverwriteAll:
      begin
        Result := True;
        OverwriteState := osAlwaysOverwrite;
      end;
    coSkip:
      begin
        Result := False;
      end;
    coSkipAll:
      begin
        Result := False;
        OverwriteState := osAlwaysSkip;
      end;
    coCancel:
      begin
        Abort;
      end;
  end;
end;

function TNamedFoldersPanel.DeleteNamedFolder(const Name: string): boolean;
begin
  Result := TNamedFolders.DeleteNamedFolder(Name);
  if Result then
    ReloadPanelContents;
end;

destructor TNamedFoldersPanel.Destroy;
begin
  SaveLastDirectory;
  inherited;
end;

function TNamedFoldersPanel.EditNamedFolder(const Name: string): boolean;
begin
  Result := TNamedFolders.EditNamedFolder(Name);
  if Result then begin
    ReloadPanelContents;
  end;
end;

function TNamedFoldersPanel.ExecuteNamedFolder(InCurrentPanel: boolean; out Handled: boolean): boolean;
var
  PanelHandle: THandle;
  Panel: TFarPanelInfo;
  Info: TFarPanelItem;
begin
  Result := False;
  Handled := True;
  if InCurrentPanel then
    PanelHandle := Self.Handle
  else
    PanelHandle := PANEL_PASSIVE;
  Panel := TFarPanelInfo.Create(PANEL_ACTIVE, True);
  try
    if Panel.CurrentItemIndex >= 0 then
      if Panel.CurrentItemIndex < Panel.Items.Count then begin
        Info := Panel.Items[Panel.CurrentItemIndex];
        if (Info.FileAttributes and FILE_ATTRIBUTE_DIRECTORY) <> 0 then begin
          Result := True;
          Handled := False;
        end
        else
          Result := TNamedFolders.SetCurrentDirToNamedFolder(Info.FileName, PanelHandle);
      end;
  finally
    FreeAndNil(Panel);
  end;
end;

function TNamedFoldersPanel.ExportNamedFolderToDirectory(const Name, TargetDirectory: string; DeleteOriginal, Silent: boolean; var OverwriteState: TOverwriteState): boolean;
begin
  ForceDirectories(TargetDirectory);
  Result := ExportNamedFolderToFile(Name, IncludeTrailingPathDelimiter(TargetDirectory) + GetLastPathElement(Name) + NAMEDFOLDERFILE_EXTENSION, DeleteOriginal, Silent, OverwriteState);
end;

function TNamedFoldersPanel.ExportNamedFolderToFile(const Name, TargetFileName: string; DeleteOriginal, Silent: boolean; var OverwriteState: TOverwriteState): boolean;
var
  Xml: TXmlDoc;
  Node: TXmlNode;
begin
  Result := False;
  Xml := TXmlDoc.Create;
  try
    Xml.SetIndentation('  ');
    Xml.Root.Name := XML_ROOTNODE;
    Node := Xml.Root.NodeNew(XML_ITEMNODE);
    if TNamedFolders.ExportNamedFolderToXml(Name, Node) then begin
      if FileExists(TargetFilename) then
        if Silent then
          Exit
        else if not AskCanOverwriteFile(TargetFileName, OverwriteState) then
          Exit;
      Xml.SaveToFile(TargetFileName);
      if DeleteOriginal then
        TNamedFolders.DeleteNamedFolder(Name);
      Result := True;
    end;
  finally
    FreeAndNil(Xml);
  end;
end;

function TNamedFoldersPanel.GetFarDirectory: PFarChar;
begin
  Result := PFarChar(fDirectory);
end;

function TNamedFoldersPanel.GetFarTitle: PFarChar;
begin
  Result := PFarChar(fTitle);
end;

function TNamedFoldersPanel.GetHandle: THandle;
begin
  Result := THandle(Self);
end;

function TNamedFoldersPanel.GetPanelItems: PPluginPanelItemArray;
begin
  ValidatePanelItems;
  if GetPanelItemsCount > 0 then
    Result := @fPanelItems[0]
  else
    Result := nil;
end;

function TNamedFoldersPanel.GetPanelItemsCount: integer;
begin
  ValidatePanelItems;
  Result := Length(fPanelItems);
end;

class function TNamedFoldersPanel.HandleToPanel(Handle: THandle; out Panel: TNamedFoldersPanel): boolean;
begin
  //Log('TNamedFoldersPanel.HandleToPanel(%x)', [Handle]);
  Result := False;
  if Handle <> 0 then begin
    Panel := TNamedFoldersPanel(Handle);
    Result := True;
  end;
  //Log('TNamedFoldersPanel.HandleToPanel -> %d', [Integer(Result)]);
end;

function TNamedFoldersPanel.ImportNamedFolderFile(const FileName, SourceDirectory: string; DeleteOriginal, Silent: boolean; var OverwriteState: TOverwriteState): boolean;
var
  Xml: TXmlDoc;
  Node: TXmlNode;
  Info, OldInfo: TNamedFolderInfo;
  CanImport: boolean;
  i: Integer;
begin
  Result := False;
  try
    Xml := TXmlDoc.Create;
    try
      Xml.LoadFromFile(IncludeTrailingPathDelimiter(SourceDirectory) + FileName);
      if Xml.Root.Name <> XML_ROOTNODE then begin
        FarMessage([MTitle, MInvalidRootNode, MOk], FMSG_WARNING, 1);
      end
      else begin
        for i := 0 to Pred(Xml.Root.NodeCount) do begin
          Node := Xml.Root.Nodes[i];
          if Node.ElementType = xeNormal then begin
            if Node.Name = XML_ITEMNODE then begin
              if TNamedFolders.ImportNamedFolderFromXml(Node, Info) then begin
                try
                  if Directory <> '' then
                    Info.Name := Directory + '/' + Info.Name;
                  CanImport := True;
                  if TNamedFolders.ReadInfoFromOptions(Info.Name, OldInfo, False) then
                    try
                      if Silent then
                        CanImport := False
                      else if OldInfo.IsDirectory then
                        CanImport := False
                      else if not AskCanOverwriteNamedFolder(OldInfo, Info, OverwriteState) then
                        CanImport := False;
                    finally
                      FreeAndNil(OldInfo);
                    end;
                  if CanImport then begin
                    TNamedFolders.SaveInfoToOptions(Info);
                    Result := True;
                  end;
                finally
                  FreeAndNil(Info);
                end;
              end;
            end;
          end;
        end;
      end;
    finally
      FreeAndNil(Xml);
    end;
  finally
    if Result then
      ReloadPanelContents;
  end;
end;

function TNamedFoldersPanel.MakeDirectory(const Name: string): boolean;
begin
  Result := TNamedFolders.MakeDirectory(GetAbsoluteDirectory(Self.Directory, Name));
  fPanelItemsValid := False;
end;

function TNamedFoldersPanel.NewNamedFolder: boolean;
begin
  Result := TNamedFolders.NewNamedFolder(Self.Directory, PANEL_PASSIVE);
  if Result then begin
    ReloadPanelContents;
  end;
end;

procedure TNamedFoldersPanel.OpenLastDirectory;
var
  Dir: string;
  Options: TOptions;
begin
  Options := TOptions.Create;
  try
    if Options.SaveLastPanelDirectory then
      Dir := Options.LastPanelDirectory
    else
      Dir := '';
    if (Dir <> '') then
      SetDirectory(Dir)
    else
      SetDirectory('/');
  finally
    FreeAndNil(Options);
  end;
end;

procedure TNamedFoldersPanel.ReloadPanelContents;
begin
  FarApi.Control(Handle, FCTL_UPDATEPANEL, 1, nil);
  FarApi.Control(Handle, FCTL_REDRAWPANEL, 0, nil);
  fPanelItemsValid := False;
end;

procedure TNamedFoldersPanel.SaveLastDirectory;
var
  Options: TOptions;
begin
  Options := TOptions.Create;
  try
    if Options.SaveLastPanelDirectory then begin
      Options.LastPanelDirectory := fDirectory;
      Options.Save;
    end;
  finally
    FreeAndNil(Options);
  end;
end;

function TNamedFoldersPanel.SelectFile(const Name: string): boolean;
var
  RedrawInfo: TPanelRedrawInfo;
  i: integer;
  NormalizedName: string;
  Panel: TFarPanelInfo;
begin
  Result := False;
  FillChar(RedrawInfo, Sizeof(RedrawInfo), 0);
  {$IFDEF FAR3_UP}
  RedrawInfo.StructSize := Sizeof(RedrawInfo);
  {$ENDIF}
  NormalizedName := TNamedFolderInfo.NormalizeName(Name);
  Panel := TFarPanelInfo.Create(Handle);
  try
    for i := 0 to Pred(Panel.Items.Count) do begin
      if AnsiCompareStr(NormalizedName, Panel.Items[i].FileName) = 0 then begin
        RedrawInfo.CurrentItem := i;
        RedrawInfo.TopPanelItem := i;
        Result := FarApi.Control(Handle, FCTL_REDRAWPANEL, 0, @RedrawInfo) <> 0;
        Break;
      end;
    end;
  finally
    FreeAndNil(Panel);
  end;
end;

procedure TNamedFoldersPanel.SetDirectory(const Value: string);
begin
  //Log('TNamedFoldersPanel.SetDirectory("%s")', [Value]);
  fDirectory := GetAbsoluteDirectory(Self.Directory, Value);
  //Log('- FDirectory = "%s"', [fDirectory]);
  fTitle := GetMsgPas(MTitle) + ': ' + fDirectory;
  fPanelItemsValid := False;
  SaveLastDirectory;
end;

procedure TNamedFoldersPanel.ValidatePanelItems;
var
  Options: TOptions;
  NameList: TStringList;
  s, FileName, DisplayFileName: string;
  DisplayNameStartIndex: integer;
  Info: TNamedFolderInfo;
  i, n: Integer;
begin
  if not fPanelItemsValid then begin
    //Log('TNamedFoldersPanel.ValidatePanelItems');
    //Log('- current directory "%s"', [Self.Directory]);
    if TNamedFolders.ReadNamedFoldersList(Self.Directory, False, NameList, DisplayNameStartIndex) then
      try
        Options := TOptions.Create;
        try
          // Add the link to parent directory, if required
          if Self.Directory <> '' then begin
            //Log('- adding parent directory');
            SetLength(fPanelItemsHelp, NameList.Count + 1);
            fPanelItemsHelp[0].FileName := '..';
            fPanelItemsHelp[0].PanelFileName := '..';
            fPanelItemsHelp[0].Description := '';
            fPanelItemsHelp[0].IsDirectory := True;
            n := 1;
          end
          else begin
            SetLength(fPanelItemsHelp, NameList.Count);
            n := 0;
          end;
          // Read info for the found names
          for i := 0 to Pred(NameList.Count) do begin
            FileName := NameList[i];
            //Log('- loading info for "%s"', [FileName]);
            s := Copy(FileName, DisplayNameStartIndex, MaxInt);
            if ExtractPathElement(s, DisplayFileName) then
              if Options.LoadNamedFolderInfo(FileName, Info) then
                try
                  //Log('- adding "%s"', [FileName]);
                  fPanelItemsHelp[n].FileName := FileName;
                  fPanelItemsHelp[n].PanelFileName := DisplayFileName;
                  fPanelItemsHelp[n].Description := Info.DisplayableDescription;
                  fPanelItemsHelp[n].IsDirectory := Info.IsDirectory;
                  Inc(n);
                finally
                  FreeAndNil(Info);
                end;
          end;
          SetLength(fPanelItemsHelp, n);
          // Step two: Prepare panel item list
          SetLength(fPanelItems, n);
          for i := 0 to Pred(n) do begin
            FillChar(fPanelItems[i], Sizeof(fPanelItems[i]), 0);
            fPanelItems[i]. {$IFDEF FAR3_UP} FileName {$ELSE} FindData.cFileName {$ENDIF} := PFarChar(fPanelItemsHelp[i].FileName);
            fPanelItems[i].Description := PFarChar(fPanelItemsHelp[i].Description);
            fPanelItems[i].Flags := {$IFDEF FAR3_UP} PPIF_NONE {$ELSE} 0 {$ENDIF} ;
            if fPanelItemsHelp[i].IsDirectory then
              fPanelItems[i]. {$IFDEF FAR3_UP} FileAttributes {$ELSE} FindData.dwFileAttributes {$ENDIF} := FILE_ATTRIBUTE_DIRECTORY
            else
              fPanelItems[i]. {$IFDEF FAR3_UP} FileAttributes {$ELSE} FindData.dwFileAttributes {$ENDIF} := FILE_ATTRIBUTE_NORMAL;
          end;
        finally
          FreeAndNil(Options);
        end;
      finally
        FreeAndNil(NameList);
      end;
    fPanelItemsValid := True;
    //Log('TNamedFoldersPanel.ValidatePanelItems -> END');
  end;
end;

end.
